/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2006-2014  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#pragma once

#include <QList>
#include <QDateTime>
#include <QtDebug>
#include <interfaces/azoth/imessage.h>

namespace LeechCraft
{
namespace Azoth
{
namespace AzothUtil
{
	namespace detail
	{
		inline IMessage* GetIMessage (IMessage *msg)
		{
			return msg;
		}

		template<typename T>
		IMessage* GetIMessage (T *msgObj)
		{
			return qobject_cast<IMessage*> (msgObj);
		}
	}

	/** @brief Standard function to purge \em messages before the given
	 * date.
	 *
	 * This function is a standard implementation to be used in
	 * LeechCraft::Azoth::ICLEntry::PurgeMessages() implementations. It
	 * deletes all the messages in the \em messages list before the given
	 * date \em before.
	 *
	 * The list of \em messages is assumed to be sorted according to the
	 * message timestamp in ascending order.
	 *
	 * If \em before is invalid all the \em messages will be erased and
	 * the list will be cleared.
	 *
	 * @param[inout] messages The list of messages to check.
	 * @param[in] before Messages with timestamp earlier than this
	 * parameter will be erased.
	 *
	 * @sa IMessage
	 */
	template<typename T>
	void StandardPurgeMessages (QList<T*>& messages, const QDateTime& before)
	{
		if (!before.isValid ())
		{
			qDeleteAll (messages);
			messages.clear ();
			return;
		}

		while (!messages.isEmpty ())
		{
			const auto msg = detail::GetIMessage (messages.at (0));
			if (!msg)
			{
				qWarning () << Q_FUNC_INFO
						<< "unable to cast"
						<< messages.at (0)
						<< "to IMessage; just blindly removing it and hoping for the best";
				messages.removeAt (0);
				continue;
			}
			if (msg->GetDateTime () < before)
				delete messages.takeAt (0);
			else
				break;
		}
	}
}
}
}
