/**********************************************************************
 * LeechCraft - modular cross-platform feature rich internet client.
 * Copyright (C) 2012  Georg Rudoy
 *
 * Boost Software License - Version 1.0 - August 17th, 2003
 *
 * Permission is hereby granted, free of charge, to any person or organization
 * obtaining a copy of the software and accompanying documentation covered by
 * this license (the "Software") to use, reproduce, display, distribute,
 * execute, and transmit the Software, and to prepare derivative works of the
 * Software, and to permit third-parties to whom the Software is furnished to
 * do so, all subject to the following:
 *
 * The copyright notices in the Software and this entire statement, including
 * the above license grant, this restriction and the following disclaimer,
 * must be included in all copies of the Software, in whole or in part, and
 * all derivative works of the Software, unless such copies or derivative
 * works are solely in the form of machine-executable object code generated by
 * a source language processor.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 * SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 * FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 **********************************************************************/

#include "artistlookup.h"
#include <memory>
#include <QUrl>
#include <QNetworkRequest>
#include <QNetworkReply>
#include <QNetworkAccessManager>
#include <QDomDocument>
#include <QDate>
#include <QtDebug>
#include <util/sll/urloperator.h>

namespace LeechCraft
{
namespace MusicZombie
{
	namespace
	{
		QString NormalizeName (QString name)
		{
			return name.remove ('!');
		}
	}

	ArtistLookup::ArtistLookup (const QString& name, QNetworkAccessManager *nam, QObject *parent)
	: QObject { parent }
	, Name_ { name }
	{
		QUrl url { "http://www.musicbrainz.org/ws/2/artist/" };
		Util::UrlOperator { url }
				("query", "artist:" + NormalizeName (name));

		auto reply = nam->get (QNetworkRequest (url));
		connect (reply,
				SIGNAL (finished ()),
				this,
				SLOT (handleFinished ()));
		connect (reply,
				SIGNAL (error (QNetworkReply::NetworkError)),
				this,
				SLOT (handleError ()));
	}

	void ArtistLookup::handleFinished ()
	{
		auto reply = qobject_cast<QNetworkReply*> (sender ());
		reply->deleteLater ();

		const auto& data = reply->readAll ();
		QDomDocument doc;
		if (!doc.setContent (data, true))
		{
			qWarning () << Q_FUNC_INFO
					<< "unable to parse reply"
					<< data;
			emit replyError ();
			return;
		}

		const auto& artists = doc.documentElement ().firstChildElement ("artist-list");
		if (artists.attribute ("count").toInt () < 1)
		{
			qWarning () << Q_FUNC_INFO
					<< "no such artists";
			emit replyError ();
			return;
		}

		QMap<int, QString> span2id;

		const auto curYear = QDate::currentDate ().year ();

		auto artist = artists.firstChildElement ("artist");
		while (!artist.isNull () && artist.attribute ("score").toInt () > 75)
		{
			std::shared_ptr<void> artistGuard { nullptr,
					[&artist] (void*) { artist = artist.nextSiblingElement ("artist"); } };

			const auto& spanElem = artist.firstChildElement ("life-span");

			const auto& beginYearElem = spanElem.firstChildElement ("begin");
			if (beginYearElem.isNull ())
				continue;

			const auto beginYear = beginYearElem.text ().simplified ().toInt ();

			const auto& endYearElem = spanElem.firstChildElement ("end");
			const auto endYear = endYearElem.isNull () ?
					curYear :
					endYearElem.text ().simplified ().toInt ();

			span2id [endYear - beginYear] = artist.attribute ("id");
		}

		qDebug () << Q_FUNC_INFO
				<< "choices for"
				<< Name_
				<< ":"
				<< span2id;

		if (span2id.isEmpty ())
			span2id [0] = artists.firstChildElement ("artist").attribute ("id");
		emit gotID (span2id.values ().last ());
	}

	void ArtistLookup::handleError ()
	{
		auto reply = qobject_cast<QNetworkReply*> (sender ());
		reply->deleteLater ();

		qWarning () << Q_FUNC_INFO
				<< Name_
				<< reply->errorString ()
				<< reply->readAll ();
		emit networkError ();
	}
}
}
